@@ -25,10 +25,11 @@ __all__ = [
     'assert_writable_and_random_access', 'prepare_rw_output_stream',
     'finalise_output',
     'DEFAULT_CHUNK_SIZE', 'chunked_write', 'chunked_digest', 'chunk_stream',
-    'ConsList', 'Singleton', 'rd', 'isoparse'
+    'ConsList', 'Singleton', 'rd', 'isoparse', 'lift_iterable_async'
 ]
 
 import pytz
+from pyhanko_certvalidator import CancelableAsyncIterator
 
 DEFAULT_CHUNK_SIZE = 4096
 """
@@ -492,3 +493,18 @@ def isoparse(dt_str: str) -> datetime:
         # assume UTC
         dt = dt.replace(tzinfo=pytz.UTC)
     return dt
+
+
+class _LiftedIterable(CancelableAsyncIterator[X]):
+    async def __anext__(self) -> X:
+        return next(self.i)
+
+    async def cancel(self):
+        return
+
+    def __init__(self, i: Iterable[X]):
+        self.i = iter(i)
+
+
+async def lift_iterable_async(i: Iterable[X]) -> CancelableAsyncIterator[X]:
+    return _LiftedIterable(i)
