@@ -12,6 +12,7 @@ from dataclasses import dataclass
 from typing import IO, Iterable, List, Optional, Tuple, Union
 
 from asn1crypto import algos, cms, keys, pem, tsp, x509
+from asn1crypto.algos import SignedDigestAlgorithm
 from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.primitives.asymmetric import padding
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
@@ -31,7 +32,9 @@ __all__ = [
     'SigningError', 'UnacceptableSignerError',
     'SignedDataCerts', 'extract_signer_info', 'extract_certificate_info',
     'load_certs_from_pemder', 'load_cert_from_pemder',
-    'load_private_key_from_pemder', 'get_pyca_cryptography_hash',
+    'load_private_key_from_pemder',
+    'load_certs_from_pemder_data', 'load_private_key_from_pemder_data',
+    'get_cms_hash_algo_for_mechanism', 'get_pyca_cryptography_hash',
     'optimal_pss_params', 'process_pss_params', 'as_signing_certificate',
     'as_signing_certificate_v2', 'match_issuer_serial',
     'check_ess_certid', 'CMSExtractionError', 'byte_range_digest',
@@ -309,11 +312,39 @@ class UnacceptableSignerError(SigningError):
     pass
 
 
-def get_pyca_cryptography_hash(algorithm, prehashed=False):
-    hash_algo = getattr(hashes, algorithm.upper())()
+def get_pyca_cryptography_hash(algorithm, prehashed=False) \
+        -> Union[hashes.HashAlgorithm, Prehashed]:
+    if algorithm.lower() == 'shake256':
+        # force the output length to 64 bytes = 512 bits
+        hash_algo = hashes.SHAKE256(digest_size=64)
+    else:
+        hash_algo = getattr(hashes, algorithm.upper())()
     return Prehashed(hash_algo) if prehashed else hash_algo
 
 
+def get_cms_hash_algo_for_mechanism(mech: SignedDigestAlgorithm) -> str:
+    """
+    Internal function that takes a :class:`.SignedDigestAlgorithm` instance
+    and returns the name of the digest algorithm that has to be used to compute
+    the ``messageDigest`` attribute.
+
+    :param mech:
+        A signature mechanism.
+    :return:
+        A digest algorithm name.
+    """
+    # anticipate resolution of this discussion:
+    # https://github.com/wbond/asn1crypto/pull/230
+
+    sig_algo = mech.signature_algo
+    if sig_algo == 'ed25519':
+        return 'sha512'
+    elif sig_algo == 'ed448':
+        return 'shake256'
+    else:
+        return mech.hash_algo
+
+
 def process_pss_params(params: algos.RSASSAPSSParams, digest_algorithm,
                        prehashed=False):
     """
@@ -399,25 +430,38 @@ def load_certs_from_pemder(cert_files):
     :return:
         A generator producing :class:`.asn1crypto.x509.Certificate` objects.
     """
-    for ca_chain_file in cert_files:
-        with open(ca_chain_file, 'rb') as f:
-            ca_chain_bytes = f.read()
-        # use the pattern from the asn1crypto docs
-        # to distinguish PEM/DER and read multiple certs
-        # from one PEM file (if necessary)
-        if pem.detect(ca_chain_bytes):
-            pems = pem.unarmor(ca_chain_bytes, multiple=True)
-            for type_name, _, der in pems:
-                if type_name is None or type_name.lower() == 'certificate':
-                    yield x509.Certificate.load(der)
-                else:  # pragma: nocover
-                    logger.debug(
-                        f'Skipping PEM block of type {type_name} in '
-                        f'CA chain file.'
-                    )
-        else:
-            # no need to unarmor, just try to load it immediately
-            yield x509.Certificate.load(ca_chain_bytes)
+    for cert_file in cert_files:
+        with open(cert_file, 'rb') as f:
+            cert_data_bytes = f.read()
+        yield from load_certs_from_pemder_data(cert_data_bytes)
+
+
+def load_certs_from_pemder_data(cert_data_bytes: bytes):
+    """
+    A convenience function to load PEM/DER-encoded certificates from
+    binary data.
+
+    :param cert_data_bytes:
+        ``bytes`` object from which to extract certificates.
+    :return:
+        A generator producing :class:`.asn1crypto.x509.Certificate` objects.
+    """
+    # use the pattern from the asn1crypto docs
+    # to distinguish PEM/DER and read multiple certs
+    # from one PEM file (if necessary)
+    if pem.detect(cert_data_bytes):
+        pems = pem.unarmor(cert_data_bytes, multiple=True)
+        for type_name, _, der in pems:
+            if type_name is None or type_name.lower() == 'certificate':
+                yield x509.Certificate.load(der)
+            else:  # pragma: nocover
+                logger.debug(
+                    f'Skipping PEM block of type {type_name} in '
+                    f'CA chain file.'
+                )
+    else:
+        # no need to unarmor, just try to load it immediately
+        yield x509.Certificate.load(cert_data_bytes)
 
 
 def load_cert_from_pemder(cert_file):
@@ -452,6 +496,22 @@ def load_private_key_from_pemder(key_file, passphrase: Optional[bytes]) \
     """
     with open(key_file, 'rb') as f:
         key_bytes = f.read()
+    return load_private_key_from_pemder_data(key_bytes, passphrase=passphrase)
+
+
+def load_private_key_from_pemder_data(
+        key_bytes: bytes,
+        passphrase: Optional[bytes]) -> keys.PrivateKeyInfo:
+    """
+    A convenience function to load PEM/DER-encoded keys from binary data.
+
+    :param key_bytes:
+        ``bytes`` object to read the key from.
+    :param passphrase:
+        Key passphrase.
+    :return:
+        A private key encoded as an unencrypted PKCS#8 PrivateKeyInfo object.
+    """
     load_fun = (
         serialization.load_pem_private_key if pem.detect(key_bytes)
         else serialization.load_der_private_key
