@@ -11,7 +11,7 @@ from datetime import datetime
 from typing import IO, List, Optional, Set, Tuple, Union
 
 import tzlocal
-from asn1crypto import cms, crl, keys, ocsp
+from asn1crypto import algos, cms, crl, keys, ocsp
 from asn1crypto import pdf as asn1_pdf
 from cryptography.hazmat.primitives import hashes
 from pyhanko_certvalidator import CertificateValidator, ValidationContext
@@ -53,6 +53,7 @@ from .cms_embedder import (
     SigObjSetup,
 )
 from .pdf_byterange import (
+    BuildProps,
     DocumentTimestamp,
     PreparedByteRangeDigest,
     SignatureObject,
@@ -293,6 +294,14 @@ class PdfSignatureMetadata:
     signing is delegated to a trusted third party).
     """
 
+    app_build_props: Optional[BuildProps] = None
+    """
+    Properties of the application that created the signature.
+    
+    If specified, this data will be recorded in the **Prop_Build**
+    dictionary of the signature.
+    """
+    
     certify: bool = False
     """
     Sign with an author (certification) signature, as opposed to an approval
@@ -987,17 +996,21 @@ class PdfSigner:
                 "algorithm specification."
             )
 
-        # TODO fall back to more useful default for weak_hash_algos
-        weak_hash_algos = (
-            signature_meta.validation_context.weak_hash_algos
+        algorithm_policy = (
+            signature_meta.validation_context.algorithm_policy
             if signature_meta.validation_context is not None
-            else ()
+            else None
         )
-        if md_algorithm in weak_hash_algos:
-            raise SigningError(
-                f"The hash algorithm {md_algorithm} is considered weak in the "
-                f"specified validation context. Please choose another."
-            )
+        if algorithm_policy is not None:
+            now = datetime.now()
+            md_algo_obj = algos.DigestAlgorithm({'algorithm': md_algorithm})
+            if not algorithm_policy\
+                    .digest_algorithm_allowed(md_algo_obj, now):
+                raise SigningError(
+                    f"The hash algorithm {md_algorithm} is not allowed in the "
+                    f"specified validation context (usage policy of type "
+                    f"{algorithm_policy.__class__.__name__})."
+                )
         return md_algorithm
 
     def register_extensions(self, pdf_out: BasePdfFileWriter, *,
@@ -1606,8 +1619,8 @@ class PdfSigningSession:
                 await validator.async_validate_usage(key_usage)
         except (PathBuildingError, PathValidationError) as e:
             raise SigningError(
-                "The signer's certificate could not be validated", e
-            )
+                "The signer's certificate could not be validated"
+            ) from e
         return signer_cert_validation_path
 
     async def _perform_prev_ts_validation(self, validation_context,
@@ -1964,6 +1977,7 @@ class PdfSigningSession:
             timestamp=system_time,
             name=name_specified if name_specified else None,
             location=signature_meta.location, reason=signature_meta.reason,
+            app_build_props=signature_meta.app_build_props,
         )
 
         # Pass in the SignatureObject settings
