@@ -6,8 +6,26 @@ from typing import IO, Iterable, List, Optional, Tuple, Type, TypeVar, Union
 from asn1crypto import cms, core, tsp, x509
 from cryptography.exceptions import InvalidSignature
 from cryptography.hazmat.primitives import hashes
-from pyhanko_certvalidator import CertificateValidator, ValidationContext
-from pyhanko_certvalidator.errors import PathBuildingError, PathValidationError
+from pyhanko_certvalidator import (
+    CancelableAsyncIterator,
+    ValidationContext,
+    find_valid_path,
+)
+from pyhanko_certvalidator.errors import (
+    DisallowedAlgorithmError,
+    ExpiredError,
+    InvalidCertificateError,
+    PathBuildingError,
+    PathValidationError,
+    RevokedError,
+    ValidationError,
+)
+from pyhanko_certvalidator.ltv.errors import TimeSlideFailure
+from pyhanko_certvalidator.path import ValidationPath
+from pyhanko_certvalidator.policy_decl import (
+    AlgorithmUsagePolicy,
+    PKIXValidationParams,
+)
 from pyhanko_certvalidator.validate import ACValidationResult, async_validate_ac
 
 from pyhanko.sign.general import (
@@ -31,12 +49,13 @@ from .status import (
     CAdESSignerAttributeAssertions,
     CertifiedAttributes,
     ClaimedAttributes,
+    RevocationDetails,
     SignatureStatus,
     StandardCMSSignatureStatus,
     TimestampSignatureStatus,
 )
 from .utils import (
-    DEFAULT_WEAK_HASH_ALGORITHMS,
+    DEFAULT_ALGORITHM_USAGE_POLICY,
     extract_message_digest,
     validate_raw,
 )
@@ -50,47 +69,59 @@ __all__ = [
     'collect_signer_attr_status', 'validate_algorithm_protection'
 ]
 
+from ...pdf_utils.misc import lift_iterable_async
+
 logger = logging.getLogger(__name__)
 
 StatusType = TypeVar('StatusType', bound=SignatureStatus)
 
 
+def _grab_signing_cert_attr(signed_attrs, v2: bool):
+    # TODO check certificate policies, enforce restrictions on chain of trust
+    # TODO document and/or mark as internal API explicitly
+    attr_name = 'signing_certificate_v2' if v2 else 'signing_certificate'
+    cls = tsp.SigningCertificateV2 if v2 else tsp.SigningCertificate
+    try:
+        value = find_unique_cms_attribute(signed_attrs, attr_name)
+        # reencode the attribute to avoid accidentally tripping the
+        # _is_mutated logic on the parent object (is important to preserve
+        # the state of the signed attributes)
+        return cls.load(value.dump())
+    except NonexistentAttributeError:
+        return None
+    except MultivaluedAttributeError as e:
+        # Banned by RFCs -> error
+        err = AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
+        raise errors.SignatureValidationError(
+            "Wrong cardinality for signing certificate attribute",
+            ades_subindication=err
+        ) from e
+
+
 def _check_signing_certificate(cert: x509.Certificate,
                                signed_attrs: cms.CMSAttributes):
     # TODO check certificate policies, enforce restrictions on chain of trust
     # TODO document and/or mark as internal API explicitly
-    def _grab(attr_name, cls):
-        try:
-            value = find_unique_cms_attribute(signed_attrs, attr_name)
-            # reencode the attribute to avoid accidentally tripping the
-            # _is_mutated logic on the parent object (is important to preserve
-            # the state of the signed attributes)
-            return cls.load(value.dump())
-        except NonexistentAttributeError:
-            return None
-        except MultivaluedAttributeError as e:
-            raise errors.SignatureValidationError(
-                "Wrong cardinality for signing certificate attribute"
-            ) from e
 
-    attr = _grab('signing_certificate_v2', tsp.SigningCertificateV2)
+    attr = _grab_signing_cert_attr(signed_attrs, v2=True)
     if attr is None:
-        attr = _grab('signing_certificate', tsp.SigningCertificate)
+        attr = _grab_signing_cert_attr(signed_attrs, v2=False)
 
     if attr is None:
         # if neither attr is present -> no constraints
         return
 
-    # we only care about the first value, the others limit the set of applicable
-    # CA certs
+    # For the main signer cert, we only care about the first value, the others
+    # limit the set of applicable CA certs
     certid = attr['certs'][0]
 
     if not check_ess_certid(cert, certid):
+        err = AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
         raise errors.SignatureValidationError(
             f"Signing certificate attribute does not match selected "
             f"signer's certificate for subject"
             f"\"{cert.subject.human_friendly}\".",
-            ades_subindication=AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
+            ades_subindication=err
         )
 
 
@@ -147,11 +178,13 @@ def validate_algorithm_protection(
             )
 
 
-def validate_sig_integrity(signer_info: cms.SignerInfo,
-                           cert: x509.Certificate,
-                           expected_content_type: str,
-                           actual_digest: bytes,
-                           weak_hash_algorithms=DEFAULT_WEAK_HASH_ALGORITHMS) \
+def validate_sig_integrity(
+        signer_info: cms.SignerInfo,
+        cert: x509.Certificate,
+        expected_content_type: str,
+        actual_digest: bytes,
+        algorithm_usage_policy: Optional[AlgorithmUsagePolicy] = None,
+        time_indic: Optional[datetime] = None) \
         -> Tuple[bool, bool]:
     """
     Validate the integrity of a signature for a particular signerInfo object
@@ -174,8 +207,10 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
         see :class:`cms.ContentType`).
     :param actual_digest:
         The actual digest to be matched to the message digest attribute.
-    :param weak_hash_algorithms:
-        List, tuple or set of weak hashing algorithms.
+    :param algorithm_usage_policy:
+        Algorithm usage policy.
+    :param time_indic:
+        Time indication for the production of the signature.
     :return:
         A tuple of two booleans. The first indicates whether the provided
         digest matches the value in the signed attributes.
@@ -186,8 +221,22 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
         signer_info['signature_algorithm']
     digest_algorithm_obj = signer_info['digest_algorithm']
     md_algorithm = digest_algorithm_obj['algorithm'].native
-    if md_algorithm in weak_hash_algorithms:
-        raise errors.WeakHashAlgorithmError(md_algorithm)
+    if algorithm_usage_policy is not None:
+        algo_allowed = algorithm_usage_policy.signature_algorithm_allowed(
+            signature_algorithm,
+            moment=time_indic,
+            public_key=cert.public_key
+        )
+        if not algo_allowed:
+            msg = (
+                f"The algorithm {signature_algorithm['algorithm'].native} "
+                f"is not allowed by the current usage policy."
+            )
+            if algo_allowed.failure_reason is not None:
+                msg += f" Reason: {algo_allowed.failure_reason}."
+            raise errors.DisallowedAlgorithmError(
+                msg, permanent=algo_allowed.not_allowed_after is None
+            )
     signature = signer_info['signature'].native
 
     signed_attrs_orig: cms.CMSAttributes = signer_info['signed_attrs']
@@ -261,7 +310,8 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
     try:
         validate_raw(
             signature, signed_data, cert, signature_algorithm, md_algorithm,
-            prehashed=prehashed, weak_hash_algorithms=weak_hash_algorithms
+            prehashed=prehashed, algorithm_policy=algorithm_usage_policy,
+            time_indic=time_indic
         )
         valid = True
     except InvalidSignature:
@@ -307,13 +357,12 @@ def extract_certs_for_validation(signed_data: cms.SignedData) \
 
 async def cms_basic_validation(
         signed_data: cms.SignedData,
-        status_cls: Type[StatusType] = SignatureStatus,
         raw_digest: bytes = None,
-        validation_context: ValidationContext = None,
+        validation_context: Optional[ValidationContext] = None,
         status_kwargs: dict = None,
-        key_usage_settings: KeyUsageConstraints = None,
-        encap_data_invalid=False):
-
+        validation_path: Optional[ValidationPath] = None,
+        pkix_validation_params: Optional[PKIXValidationParams] = None,
+        *, key_usage_settings: KeyUsageConstraints):
     """
     Perform basic validation of CMS and PKCS#7 signatures in isolation
     (i.e. integrity and trust checks).
@@ -325,11 +374,14 @@ async def cms_basic_validation(
     cert = cert_info.signer_cert
     other_certs = cert_info.other_certs
 
-    weak_hash_algos = None
+    algorithm_policy = None
+    time_indic = None
     if validation_context is not None:
-        weak_hash_algos = validation_context.weak_hash_algos
-    if weak_hash_algos is None:
-        weak_hash_algos = DEFAULT_WEAK_HASH_ALGORITHMS
+        algorithm_policy = validation_context.algorithm_policy
+        time_indic = validation_context.best_signature_time
+    validation_context = validation_context or ValidationContext()
+    if algorithm_policy is None:
+        algorithm_policy = DEFAULT_ALGORITHM_USAGE_POLICY
 
     signature_algorithm: cms.SignedDigestAlgorithm = \
         signer_info['signature_algorithm']
@@ -346,31 +398,38 @@ async def cms_basic_validation(
         raw_digest = md.finalize()
 
     # first, do the cryptographic identity checks
+    # TODO theoretically (e.g. DSA with param inheritance) this requires
+    #  doing the X.509 validation step first. Since nobody cares about DSA
+    #  (let alone DSA with inherited parameters), that's just a "nice to have".
     try:
         intact, valid = validate_sig_integrity(
             signer_info, cert, expected_content_type=expected_content_type,
-            actual_digest=raw_digest, weak_hash_algorithms=weak_hash_algos
+            actual_digest=raw_digest, algorithm_usage_policy=algorithm_policy,
+            time_indic=time_indic
         )
     except CMSStructuralError as e:
         raise errors.SignatureValidationError(
-            "CMS structural error: " + e.failure_message
+            "CMS structural error: " + e.failure_message,
+            ades_subindication=AdESFailure.FORMAT_FAILURE
         ) from e
 
-    # if the data being encapsulated by the signature is itself invalid,
-    #  this flag is set
-    intact &= not encap_data_invalid
-    valid &= intact
-
     # next, validate trust
-    ades_status = path = None
+    ades_status = path = revo_details = None
     if valid:
         try:
-            validator = CertificateValidator(
-                cert, intermediate_certs=other_certs,
-                validation_context=validation_context
-            )
-            ades_status, path = await status_cls.validate_cert_usage(
-                validator, key_usage_settings=key_usage_settings
+            validation_context.certificate_registry\
+                .register_multiple(other_certs)
+
+            if validation_path is not None:
+                paths = lift_iterable_async([validation_path])
+            else:
+                paths = validation_context.path_builder\
+                    .async_build_paths_lazy(cert)
+
+            ades_status, revo_details, path = await validate_cert_usage(
+                cert, validation_context,
+                key_usage_settings=key_usage_settings,
+                paths=paths, pkix_validation_params=pkix_validation_params,
             )
         except ValueError as e:
             logger.error("Processing error in validation process", exc_info=e)
@@ -380,19 +439,46 @@ async def cms_basic_validation(
     status_kwargs.update(
         intact=intact, valid=valid, signing_cert=cert,
         md_algorithm=md_algorithm, pkcs7_signature_mechanism=mechanism,
-        trust_problem_indic=ades_status, validation_path=path
+        trust_problem_indic=ades_status, validation_path=path,
+        revocation_details=revo_details
     )
     return status_kwargs
 
 
+async def validate_cert_usage(
+        cert: x509.Certificate,
+        validation_context: ValidationContext,
+        key_usage_settings: KeyUsageConstraints,
+        paths: CancelableAsyncIterator[ValidationPath],
+        pkix_validation_params: Optional[PKIXValidationParams] = None):
+    """
+    Low-level certificate validation routine.
+    Internal API.
+    """
+
+    async def _check():
+        key_usage_settings.validate(cert)
+        return await find_valid_path(
+            cert, paths,
+            validation_context=validation_context,
+            pkix_validation_params=pkix_validation_params
+        )
+    # validate usage without going through pyhanko_certvalidator
+    ades_status, revo_details, path = \
+        await handle_certvalidator_errors(_check())
+    if ades_status is not None:
+        subj = cert.subject.human_friendly
+        logger.warning(f"Chain of trust validation for {subj} failed.")
+    return ades_status, revo_details, path
+
+
 async def async_validate_cms_signature(
                            signed_data: cms.SignedData,
                            status_cls: Type[StatusType] = SignatureStatus,
                            raw_digest: bytes = None,
                            validation_context: ValidationContext = None,
                            status_kwargs: dict = None,
-                           key_usage_settings: KeyUsageConstraints = None,
-                           encap_data_invalid=False):
+                           key_usage_settings: KeyUsageConstraints = None):
     """
     Validate a CMS signature (i.e. a ``SignedData`` object).
 
@@ -410,20 +496,14 @@ async def async_validate_cms_signature(
     :param key_usage_settings:
         A :class:`.KeyUsageConstraints` object specifying which key usages
         must or must not be present in the signer's certificate.
-    :param encap_data_invalid:
-        If ``True``, the encapsulated data inside the CMS is invalid,
-        but the remaining validation logic still has to be run (e.g. a
-        timestamp token, which requires validation of the embedded message
-        imprint).
-
-        This option is considered internal API, the semantics of which may
-        change without notice in the future.
     :return:
         A :class:`.SignatureStatus` object (or an instance of a proper subclass)
     """
+    key_usage_settings = \
+        status_cls.default_usage_constraints(key_usage_settings)
     status_kwargs = await cms_basic_validation(
-        signed_data, status_cls, raw_digest, validation_context,
-        status_kwargs, key_usage_settings, encap_data_invalid
+        signed_data, raw_digest, validation_context,
+        status_kwargs, key_usage_settings=key_usage_settings
     )
     return status_cls(**status_kwargs)
 
@@ -448,7 +528,8 @@ def extract_self_reported_ts(signer_info: cms.SignerInfo) -> Optional[datetime]:
         pass
 
 
-def extract_tst_data(signer_info, signed=False) -> Optional[cms.SignedData]:
+def extract_tst_data(signer_info: cms.SignerInfo, signed: bool = False) \
+        -> Optional[cms.SignedData]:
     """
     Extract signed data associated with a timestamp token.
 
@@ -556,7 +637,7 @@ async def collect_timing_info(signer_info: cms.SignerInfo,
 
 async def validate_tst_signed_data(
         tst_signed_data: cms.SignedData,
-        validation_context: ValidationContext,
+        validation_context: Optional[ValidationContext],
         expected_tst_imprint: bytes):
     """
     Validate the ``SignedData`` of a time stamp token.
@@ -579,8 +660,17 @@ async def validate_tst_signed_data(
         tst_info = tst_info_bytes.parsed
     if not isinstance(tst_info, tsp.TSTInfo):
         raise errors.SignatureValidationError(
-            "SignedData does not encapsulate TSTInfo"
+            "SignedData does not encapsulate TSTInfo",
+            ades_subindication=AdESFailure.FORMAT_FAILURE
         )
+    timestamp = tst_info['gen_time'].native
+
+    ku_settings = TimestampSignatureStatus.default_usage_constraints()
+    status_kwargs = await cms_basic_validation(
+        tst_signed_data, validation_context=validation_context,
+        status_kwargs={'timestamp': timestamp},
+        key_usage_settings=ku_settings
+    )
     # compare the expected TST digest against the message imprint
     # inside the signed data
     tst_imprint = tst_info['message_imprint']['hashed_message'].native
@@ -589,16 +679,8 @@ async def validate_tst_signed_data(
             f"Timestamp token imprint is {tst_imprint.hex()}, but expected "
             f"{expected_tst_imprint.hex()}."
         )
-        encap_data_invalid = True
-    else:
-        encap_data_invalid = False
-    timestamp = tst_info['gen_time'].native
-    return await cms_basic_validation(
-        tst_signed_data, status_cls=TimestampSignatureStatus,
-        validation_context=validation_context,
-        status_kwargs={'timestamp': timestamp},
-        encap_data_invalid=encap_data_invalid
-    )
+        status_kwargs['intact'] = False
+    return status_kwargs
 
 
 async def process_certified_attrs(
@@ -632,7 +714,10 @@ async def collect_signer_attr_status(
     except NonexistentAttributeError:
         signer_attrs = None
     except MultivaluedAttributeError as e:
-        raise errors.SignatureValidationError(str(e)) from e
+        # TODO downgrade to a warning?
+        raise errors.SignatureValidationError(
+            str(e), ades_subindication=AdESFailure.FORMAT_FAILURE
+        ) from e
 
     result = {}
     cades_ac_results = None
@@ -781,9 +866,10 @@ async def async_validate_detached_cms(
         signer_info, ts_validation_context=ts_validation_context,
         raw_digest=digest_bytes
     )
+    key_usage_settings = \
+        StandardCMSSignatureStatus.default_usage_constraints(key_usage_settings)
     status_kwargs = await cms_basic_validation(
-        signed_data, status_cls=StandardCMSSignatureStatus,
-        raw_digest=digest_bytes,
+        signed_data, raw_digest=digest_bytes,
         validation_context=signer_validation_context,
         status_kwargs=status_kwargs,
         key_usage_settings=key_usage_settings
@@ -802,3 +888,68 @@ async def async_validate_detached_cms(
         )
     )
     return StandardCMSSignatureStatus(**status_kwargs)
+
+
+async def handle_certvalidator_errors(coro):
+    """
+    Internal error handling function that maps certvalidator errors
+    to AdES status indications.
+
+    :param coro:
+    :return:
+    """
+    revo_details = None
+    try:
+        return None, None, await coro
+    except InvalidCertificateError as e:
+        logger.warning(e.failure_msg, exc_info=e)
+        ades_status = AdESIndeterminate.CHAIN_CONSTRAINTS_FAILURE
+    except TimeSlideFailure as e:
+        logger.warning(e.failure_msg, exc_info=e)
+        ades_status = AdESIndeterminate.NO_POE
+    except DisallowedAlgorithmError as e:
+        if e.banned_since is None:
+            # permaban
+            ades_status = AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE
+        else:
+            # could get resolved with more POEs
+            ades_status = AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE_NO_POE
+    except RevokedError as e:
+        logger.warning(e.failure_msg)
+        if e.is_side_validation:
+            # don't report this as a revocation event
+            ades_status = AdESIndeterminate.CERTIFICATE_CHAIN_GENERAL_FAILURE
+        elif e.is_ee_cert:
+            ades_status = AdESIndeterminate.REVOKED_NO_POE
+            revo_details = RevocationDetails(
+                ca_revoked=False, revocation_date=e.revocation_dt,
+                revocation_reason=e.reason
+            )
+        else:
+            ades_status = AdESIndeterminate.REVOKED_CA_NO_POE
+            revo_details = RevocationDetails(
+                ca_revoked=True, revocation_date=e.revocation_dt,
+                revocation_reason=e.reason
+            )
+    except PathBuildingError as e:
+        logger.warning("Failed to build path", exc_info=e)
+        ades_status = AdESIndeterminate.NO_CERTIFICATE_CHAIN_FOUND
+    except ExpiredError as e:
+        logger.warning(e.failure_msg)
+        if not e.is_side_validation and e.is_ee_cert:
+            # TODO modify certvalidator to perform revinfo checks on
+            #  expired certs, possibly as an option. If this happens, we
+            #  can potentially emit the more accurate status
+            #  OUT_OF_BOUNDS_NOT_REVOKED here in cases where it applies.
+            ades_status = AdESIndeterminate.OUT_OF_BOUNDS_NO_POE
+        else:
+            ades_status = AdESIndeterminate.CERTIFICATE_CHAIN_GENERAL_FAILURE
+    except PathValidationError as e:
+        logger.warning(e.failure_msg, exc_info=e)
+        # TODO verify whether this is appropriate
+        ades_status = AdESIndeterminate.CHAIN_CONSTRAINTS_FAILURE
+    except ValidationError as e:
+        logger.warning(e.failure_msg, exc_info=e)
+        ades_status = AdESIndeterminate.CERTIFICATE_CHAIN_GENERAL_FAILURE
+
+    return ades_status, revo_details, None
