@@ -6,7 +6,7 @@ from typing import Iterator, Optional, TypeVar
 from asn1crypto import cms
 from asn1crypto import pdf as asn1_pdf
 from pyhanko_certvalidator import ValidationContext
-from pyhanko_certvalidator.context import (
+from pyhanko_certvalidator.policy_decl import (
     CertRevTrustPolicy,
     RevocationCheckingPolicy,
     RevocationCheckingRule,
@@ -82,21 +82,32 @@ class RevocationInfoValidationType(Enum):
         return tuple(m.value for m in cls)
 
 
-DEFAULT_LTV_INTERNAL_REVO_POLICY = CertRevTrustPolicy(
-    RevocationCheckingPolicy(
-        ee_certificate_rule=RevocationCheckingRule.CHECK_IF_DECLARED,
-        intermediate_ca_cert_rule=RevocationCheckingRule.CHECK_IF_DECLARED,
-    )
+DEFAULT_LTV_INTERNAL_REVO_CHECK_POLICY = RevocationCheckingPolicy(
+    ee_certificate_rule=RevocationCheckingRule.CHECK_IF_DECLARED,
+    intermediate_ca_cert_rule=RevocationCheckingRule.CHECK_IF_DECLARED,
 )
 
-STRICT_LTV_INTERNAL_REVO_POLICY = CertRevTrustPolicy(
-    RevocationCheckingPolicy(
-        ee_certificate_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
-        intermediate_ca_cert_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
-    ),
+
+def _default_ltv_internal_revo_policy(**kwargs):
+    return CertRevTrustPolicy(
+        revocation_checking_policy=DEFAULT_LTV_INTERNAL_REVO_CHECK_POLICY,
+        **kwargs
+    )
+
+
+STRICT_LTV_INTERNAL_REVO_CHECK_POLICY = RevocationCheckingPolicy(
+    ee_certificate_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
+    intermediate_ca_cert_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
 )
 
 
+def _strict_ltv_internal_revo_policy(**kwargs):
+    return CertRevTrustPolicy(
+        revocation_checking_policy=STRICT_LTV_INTERNAL_REVO_CHECK_POLICY,
+        **kwargs
+    )
+
+
 def _strict_vc_context_kwargs(timestamp, validation_context_kwargs):
     # create a new validation context using the timestamp value as the time
     # of evaluation, turn off fetching and load OCSP responses / CRL data
@@ -109,6 +120,7 @@ def _strict_vc_context_kwargs(timestamp, validation_context_kwargs):
 
     revinfo_policy: CertRevTrustPolicy \
         = validation_context_kwargs.get('revinfo_policy', None)
+    retroactive = validation_context_kwargs.get('retroactive_revinfo', False)
     if revinfo_policy is None:
         # handle legacy revocation mode
         legacy_rm = validation_context_kwargs.pop('revocation_mode', None)
@@ -118,10 +130,14 @@ def _strict_vc_context_kwargs(timestamp, validation_context_kwargs):
             )
         elif legacy_rm == 'soft-fail':
             # fall back to the default
-            revinfo_policy = DEFAULT_LTV_INTERNAL_REVO_POLICY
+            revinfo_policy = _default_ltv_internal_revo_policy(
+                retroactive_revinfo=retroactive
+            )
     elif not revinfo_policy.revocation_checking_policy.essential:
         # also in this case, we sub in the default
-        revinfo_policy = DEFAULT_LTV_INTERNAL_REVO_POLICY
+        revinfo_policy = _strict_ltv_internal_revo_policy(
+            retroactive_revinfo=retroactive
+        )
 
     validation_context_kwargs['revinfo_policy'] = revinfo_policy
 
@@ -191,7 +207,7 @@ class _TimestampTrustData:
     latest_dts: EmbeddedPdfSignature
     earliest_ts_status: TimestampSignatureStatus
     ts_chain_length: int
-    current_signature_vc: ValidationContext
+    current_signature_vc_kwargs: dict
 
 
 def _instantiate_ltv_vc(emb_timestamp: EmbeddedPdfSignature,
@@ -204,6 +220,8 @@ def _instantiate_ltv_vc(emb_timestamp: EmbeddedPdfSignature,
         )
         return dss.as_validation_context(validation_context_kwargs)
     except NoDSSFoundError:
+        validation_context_kwargs.setdefault('crls', ())
+        validation_context_kwargs.setdefault('ocsps', ())
         return ValidationContext(**validation_context_kwargs)
 
 
@@ -237,7 +255,8 @@ async def _establish_timestamp_trust_lta(
 
     return _TimestampTrustData(
         latest_dts=emb_timestamp, earliest_ts_status=ts_status,
-        ts_chain_length=ts_count + 1, current_signature_vc=current_vc,
+        ts_chain_length=ts_count + 1,
+        current_signature_vc_kwargs=validation_context_kwargs
     )
 
 
@@ -304,26 +323,31 @@ async def async_validate_pdf_ltv_signature(
     # To validate the first timestamp, allow fetching by default
     # we'll turn it off later
     validation_context_kwargs.setdefault('allow_fetching', True)
+    validation_context_kwargs.setdefault('retroactive_revinfo', False)
     # same for revocation_mode: if force_revinfo is false, we simply turn on
     # hard-fail by default for now. Once the timestamp is validated,
     # we switch to hard-fail forcibly.
+    retroactive = validation_context_kwargs['retroactive_revinfo']
     if force_revinfo:
         validation_context_kwargs['revinfo_policy'] \
-            = STRICT_LTV_INTERNAL_REVO_POLICY
+            = _default_ltv_internal_revo_policy(retroactive_revinfo=retroactive)
         if ac_validation_context_kwargs is not None:
             ac_validation_context_kwargs['revinfo_policy'] \
-                = STRICT_LTV_INTERNAL_REVO_POLICY
+                = _strict_ltv_internal_revo_policy(
+                    retroactive_revinfo=retroactive
+                )
     elif 'revocation_mode' not in validation_context_kwargs:
         validation_context_kwargs.setdefault(
             'revinfo_policy',
-            DEFAULT_LTV_INTERNAL_REVO_POLICY
+            _default_ltv_internal_revo_policy(retroactive_revinfo=retroactive)
         )
         if ac_validation_context_kwargs is not None:
             ac_validation_context_kwargs.setdefault(
                 'revinfo_policy',
-                DEFAULT_LTV_INTERNAL_REVO_POLICY
+                _default_ltv_internal_revo_policy(retroactive_revinfo=retroactive)
             )
 
+    current_vc: ValidationContext
     reader = embedded_sig.reader
     if validation_type == RevocationInfoValidationType.ADOBE_STYLE:
         dss = None
@@ -340,8 +364,9 @@ async def async_validate_pdf_ltv_signature(
         else:
             current_vc = bootstrap_validation_context
             # add the certs from the DSS
-            for cert in dss.load_certs():
-                current_vc.certificate_registry.add_other_cert(cert)
+            current_vc.certificate_registry.register_multiple(
+                dss.load_certs()
+            )
 
     embedded_sig.compute_digest()
     embedded_sig.compute_tst_digest()
@@ -362,7 +387,12 @@ async def async_validate_pdf_ltv_signature(
             validation_context_kwargs=validation_context_kwargs,
             until_revision=embedded_sig.signed_revision
         )
-        current_vc = ts_trust_data.current_signature_vc
+        validation_context_kwargs = ts_trust_data.current_signature_vc_kwargs
+        # if we found a timestamp, use it to pass to the next VC iteration
+        if ts_trust_data.latest_dts:
+            current_vc = _instantiate_ltv_vc(
+                ts_trust_data.latest_dts, validation_context_kwargs
+            )
         # In PAdES-LTA, we should only rely on DSS information that is covered
         # by an appropriate document timestamp.
         # If the validation profile is PAdES-LTA, then we must have seen
@@ -387,6 +417,11 @@ async def async_validate_pdf_ltv_signature(
         earliest_good_timestamp_st = await establish_timestamp_trust(
             tst_signed_data, current_vc, embedded_sig.tst_signature_digest
         )
+        # update validation context moments to the new signature POE
+        signature_poe = earliest_good_timestamp_st.timestamp
+        validation_context_kwargs['moment'] = signature_poe
+        if ac_validation_context_kwargs is not None:
+            ac_validation_context_kwargs['moment'] = signature_poe
     elif validation_type == RevocationInfoValidationType.PADES_LTA \
             and ts_trust_data.ts_chain_length == 1:
         # TODO Pretty sure that this is the spirit of the LTA profile,
@@ -434,14 +469,16 @@ async def async_validate_pdf_ltv_signature(
             )
     else:
         # in the LTA profile, we should use only DSS information covered
-        # by the last relevant timestamp, so the correct VC is current_vc
-        current_vc.moment = earliest_good_timestamp_st.timestamp
-        stored_vc = current_vc
+        # by the last relevant timestamp, so the correct VC is encoded
+        # by the current validation context kwargs retrieved by the lta helper
+        # logic (note that we updated the POE timestamps further up)
+        stored_vc = _instantiate_ltv_vc(
+            ts_trust_data.latest_dts, validation_context_kwargs
+        )
         if ac_validation_context_kwargs is not None:
             stored_ac_vc = _instantiate_ltv_vc(
                 ts_trust_data.latest_dts, ac_validation_context_kwargs
             )
-            stored_ac_vc.moment = earliest_good_timestamp_st.timestamp
 
     # Now, we evaluate the validity of the timestamp guaranteeing the signature
     #  *within* the LTV context.
@@ -482,8 +519,10 @@ async def async_validate_pdf_ltv_signature(
         'signer_reported_dt': earliest_good_timestamp_st.timestamp,
         'timestamp_validity': timestamp_status
     })
+    key_usage_settings = \
+        PdfSignatureStatus.default_usage_constraints(key_usage_settings)
     status_kwargs = await cms_basic_validation(
-        embedded_sig.signed_data, status_cls=PdfSignatureStatus,
+        embedded_sig.signed_data,
         raw_digest=embedded_sig.external_digest,
         validation_context=stored_vc, status_kwargs=status_kwargs,
         key_usage_settings=key_usage_settings
