@@ -5,19 +5,13 @@ from datetime import datetime
 from enum import unique
 from typing import ClassVar, Collection, Dict, Iterable, Optional, Set, Union
 
-from asn1crypto import cms, core, keys, x509
-from pyhanko_certvalidator import CertificateValidator
-from pyhanko_certvalidator.errors import (
-    InvalidCertificateError,
-    PathBuildingError,
-    PathValidationError,
-    RevokedError,
-)
+from asn1crypto import cms, core, crl, keys, x509
+from pyhanko_certvalidator.errors import PathBuildingError, PathValidationError
 from pyhanko_certvalidator.path import ValidationPath
 from pyhanko_certvalidator.validate import ACValidationResult
 
 from ...pdf_utils.misc import OrderedEnum
-from ..ades.report import AdESFailure, AdESIndeterminate, AdESSubIndic
+from ..ades.report import AdESFailure, AdESSubIndic
 from ..diff_analysis import (
     DiffResult,
     ModificationLevel,
@@ -33,12 +27,36 @@ __all__ = [
     'CAdESSignerAttributeAssertions',
     'StandardCMSSignatureStatus',
     'SignatureCoverageLevel', 'ModificationInfo',
-    'PdfSignatureStatus', 'DocumentTimestampStatus'
+    'PdfSignatureStatus', 'DocumentTimestampStatus',
+    'RevocationDetails', 'SignerAttributeStatus'
 ]
 
 logger = logging.getLogger(__name__)
 
 
+@dataclass(frozen=True)
+class RevocationDetails:
+    """
+    Contains details about a certificate revocation related to a signature.
+    """
+
+    ca_revoked: bool
+    """
+    If ``False``, the revoked certificate is the signer's. If ``True``, there's
+    a revoked CA certificate higher up the chain.
+    """
+
+    revocation_date: datetime
+    """
+    The date and time of revocation.
+    """
+
+    revocation_reason: crl.CRLReason
+    """
+    The reason why the certificate was revoked.
+    """
+
+
 @dataclass(frozen=True)
 class SignatureStatus:
     """
@@ -50,12 +68,15 @@ class SignatureStatus:
     Reports whether the signature is *intact*, i.e. whether the hash of the
     message content (which may or may not be embedded inside the CMS object
     itself) matches the hash value that was signed.
+
+    If there are no signed attributes, this is equal to :attr:`valid`.
     """
 
     valid: bool
     """
-    Reports whether the signature is *valid*, i.e. whether the hash's signature
-    actually validates.
+    Reports whether the signature is *valid*, i.e. whether the signature
+    in the CMS object itself (usually computed over a hash of the signed
+    attributes) is cryptographically valid.
     """
 
     trust_problem_indic: Optional[AdESSubIndic]
@@ -86,6 +107,12 @@ class SignatureStatus:
     certificate to a trusted root certificate.
     """
 
+    revocation_details: Optional[RevocationDetails]
+    """
+    Details on why and when the signer's certificate (or another certificate
+    in the chain) was revoked.
+    """
+
     # XXX frozenset makes more sense here, but asn1crypto doesn't allow that
     #  (probably legacy behaviour)
     key_usage: ClassVar[Set[str]] = {'non_repudiation'}
@@ -118,11 +145,7 @@ class SignatureStatus:
         If this field is ``True``, then obviously :attr:`trusted` will be
         ``False``.
         """
-        return self.trust_problem_indic in (
-            AdESFailure.REVOKED,
-            AdESIndeterminate.REVOKED_CA_NO_POE,
-            AdESIndeterminate.REVOKED_NO_POE
-        )
+        return self.revocation_details is not None
 
     @property
     def trusted(self) -> bool:
@@ -130,7 +153,7 @@ class SignatureStatus:
         Reports whether the signer's certificate is trusted w.r.t. the currently
         relevant validation context and key usage requirements.
         """
-        return self.valid and self.trust_problem_indic is None
+        return self.valid and self.intact and self.trust_problem_indic is None
 
     # TODO explain in more detail.
     def summary(self, delimiter=','):
@@ -143,9 +166,10 @@ class SignatureStatus:
             return 'INVALID'
 
     @classmethod
-    async def validate_cert_usage(
-            cls, validator: CertificateValidator,
-            key_usage_settings: KeyUsageConstraints = None):
+    def default_usage_constraints(
+                cls, key_usage_settings: Optional[KeyUsageConstraints] = None) \
+            -> KeyUsageConstraints:
+
         key_usage_settings = key_usage_settings or KeyUsageConstraints()
         key_usage_settings = KeyUsageConstraints(
             key_usage=(
@@ -157,37 +181,7 @@ class SignatureStatus:
                 else key_usage_settings.extd_key_usage
             )
         )
-        cert: x509.Certificate = validator._certificate
-
-        path = None
-        try:
-            # validate usage without going through pyhanko_certvalidator
-            key_usage_settings.validate(cert)
-            path = await validator.async_validate_usage(key_usage=set())
-            ades_status = None
-        except InvalidCertificateError as e:
-            # TODO accumulate these somewhere?
-            logger.warning(e)
-            ades_status = AdESIndeterminate.CHAIN_CONSTRAINTS_FAILURE
-        except RevokedError:
-            # TODO have certvalidator report with which of the three
-            #  revocation cases under AdES we're dealing with.
-            # (For now, assume the worst.)
-            ades_status = AdESFailure.REVOKED
-        except PathBuildingError as e:
-            logger.warning(e)
-            ades_status = AdESIndeterminate.NO_CERTIFICATE_CHAIN_FOUND
-        except PathValidationError as e:
-            logger.warning(e)
-            # TODO make error reporting in certvalidator more granular
-            #  so we can actually set proper AdES values
-            #  (e.g. expiration-related ones, distinguish between CA and EE,
-            #   revinfo freshness...)
-            ades_status = AdESIndeterminate.CERTIFICATE_CHAIN_GENERAL_FAILURE
-        if ades_status is not None:
-            subj = cert.subject.human_friendly
-            logger.warning(f"Chain of trust validation for {subj} failed.")
-        return ades_status, path
+        return key_usage_settings
 
     @property
     def _trust_anchor(self):
@@ -434,30 +428,7 @@ class CAdESSignerAttributeAssertions:
 
 
 @dataclass(frozen=True)
-class StandardCMSSignatureStatus(SignatureStatus):
-    """
-    Status of a standard "end-entity" CMS signature, potentially with
-    timing information embedded inside.
-    """
-
-    signer_reported_dt: Optional[datetime] = None
-    """
-    Signer-reported signing time, if present in the signature.
-
-    Generally speaking, this timestamp should not be taken as fact.
-    """
-
-    timestamp_validity: Optional[TimestampSignatureStatus] = None
-    """
-    Validation status of the signature timestamp token embedded in this
-    signature, if present.
-    """
-
-    content_timestamp_validity: Optional[TimestampSignatureStatus] = None
-    """
-    Validation status of the content timestamp token embedded in this
-    signature, if present.
-    """
+class SignerAttributeStatus:
 
     ac_attrs: Optional[CertifiedAttributes] = None
     """
@@ -496,6 +467,33 @@ class StandardCMSSignatureStatus(SignatureStatus):
     attribute defined in CAdES.
     """
 
+
+@dataclass(frozen=True)
+class StandardCMSSignatureStatus(SignerAttributeStatus, SignatureStatus):
+    """
+    Status of a standard "end-entity" CMS signature, potentially with
+    timing information embedded inside.
+    """
+
+    signer_reported_dt: Optional[datetime] = None
+    """
+    Signer-reported signing time, if present in the signature.
+
+    Generally speaking, this timestamp should not be taken as fact.
+    """
+
+    timestamp_validity: Optional[TimestampSignatureStatus] = None
+    """
+    Validation status of the signature timestamp token embedded in this
+    signature, if present.
+    """
+
+    content_timestamp_validity: Optional[TimestampSignatureStatus] = None
+    """
+    Validation status of the content timestamp token embedded in this
+    signature, if present.
+    """
+
     @property
     def bottom_line(self) -> bool:
         """
@@ -512,13 +510,14 @@ class StandardCMSSignatureStatus(SignatureStatus):
         if ts is None:
             timestamp_ok = True
         else:
-            timestamp_ok = ts.valid and ts.trusted
+            timestamp_ok = ts.valid and ts.intact and ts.trusted
 
         content_ts = self.content_timestamp_validity
         if content_ts is None:
             content_timestamp_ok = True
         else:
-            content_timestamp_ok = content_ts.valid and content_ts.trusted
+            content_timestamp_ok = \
+                content_ts.valid and content_ts.intact and content_ts.trusted
 
         return (
                 self.intact and self.valid and self.trusted and timestamp_ok
