@@ -1,3 +1,6 @@
+from datetime import datetime
+from typing import Optional
+
 from asn1crypto import cms, x509
 from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.asymmetric import padding
@@ -9,7 +12,12 @@ from cryptography.hazmat.primitives.asymmetric.ec import (
 from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
 from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
+from pyhanko_certvalidator.policy_decl import (
+    AlgorithmUsagePolicy,
+    DisallowWeakAlgorithmsPolicy,
+)
 
+from ..ades.report import AdESIndeterminate
 from ..general import (
     MultivaluedAttributeError,
     NonexistentAttributeError,
@@ -17,27 +25,57 @@ from ..general import (
     get_pyca_cryptography_hash,
     process_pss_params,
 )
-from .errors import SignatureValidationError, WeakHashAlgorithmError
+from .errors import DisallowedAlgorithmError, SignatureValidationError
 
 DEFAULT_WEAK_HASH_ALGORITHMS = frozenset({'sha1', 'md5', 'md2'})
 
+DEFAULT_ALGORITHM_USAGE_POLICY = \
+    DisallowWeakAlgorithmsPolicy(DEFAULT_WEAK_HASH_ALGORITHMS)
 
-def validate_raw(signature: bytes, signed_data: bytes, cert: x509.Certificate,
-                 signature_algorithm: cms.SignedDigestAlgorithm,
-                 md_algorithm: str, prehashed=False,
-                 weak_hash_algorithms=DEFAULT_WEAK_HASH_ALGORITHMS):
+
+def validate_raw(
+        signature: bytes, signed_data: bytes, cert: x509.Certificate,
+        signature_algorithm: cms.SignedDigestAlgorithm,
+        md_algorithm: str, prehashed=False,
+        algorithm_policy: Optional[AlgorithmUsagePolicy]
+        = DEFAULT_ALGORITHM_USAGE_POLICY,
+        time_indic: Optional[datetime] = None):
     """
     Validate a raw signature. Internal API.
     """
-    try:
-        sig_md_algorithm = signature_algorithm.hash_algo
-    except (ValueError, AttributeError):
-        sig_md_algorithm = None
+    if algorithm_policy is not None:
+        sig_algo_allowed = algorithm_policy.signature_algorithm_allowed(
+            signature_algorithm, moment=time_indic,
+            public_key=cert.public_key
+        )
+        if not sig_algo_allowed:
+            msg = (
+                f"Signature algorithm "
+                f"{signature_algorithm['algorithm'].native} is not allowed "
+                f"by the current usage policy."
+            )
+            if sig_algo_allowed.failure_reason is not None:
+                msg += f" Reason: {sig_algo_allowed.failure_reason}"
+            raise DisallowedAlgorithmError(
+                msg, permanent=sig_algo_allowed.not_allowed_after is None
+            )
 
-    if sig_md_algorithm is not None:
-        if sig_md_algorithm in weak_hash_algorithms:
-            raise WeakHashAlgorithmError(md_algorithm)
-        md_algorithm = sig_md_algorithm.upper()
+    try:
+        sig_hash_algo = cms.DigestAlgorithm(
+            {'algorithm': signature_algorithm.hash_algo}
+        )
+    except ValueError:
+        sig_hash_algo = None
+
+    hash_algo_obj = cms.DigestAlgorithm({'algorithm': md_algorithm})
+    if sig_hash_algo is not None \
+            and sig_hash_algo.dump() != hash_algo_obj.dump():
+        raise SignatureValidationError(
+            f"Digest algorithm {hash_algo_obj['algorithm'].native} does not "
+            f"match value implied by signature algorithm "
+            f"{signature_algorithm['algorithm'].native}",
+            ades_subindication=AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE
+        )
 
     verify_md = get_pyca_cryptography_hash(md_algorithm, prehashed=prehashed)
 
